# -*- coding: utf-8 -*-
# SPDX-License-Identifier: BSD-3-Clause
# Author: Tom Dupre la Tour (original work)
#         Roman Feldbauer (adaptions for scikit-hubness)
#         Daniel Obraczka (adaptions for kiez)
# PEP 563: Postponed Evaluation of Annotations

from __future__ import annotations

import logging
from typing import Tuple

import numpy as np
from tqdm.auto import tqdm

from kiez.io.temp_file_handling import create_tempfile_preferably_in_dir
from kiez.neighbors.neighbor_algorithm_base import NNAlgorithmWithJoblib

try:
    import annoy  # noqa: autoimport

except ImportError:  # pragma: no cover
    annoy = None


class Annoy(NNAlgorithmWithJoblib):
    """
    Wrapper for Spotify's approximate nearest neighbor library

    Parameters
    ----------
    n_candidates: int
        number of nearest neighbors used in search
    metric: str, default = 'euclidean'
        distance measure used in search
        possible measures are found in :obj:`Annoy.valid_metrics`
    n_trees: int, default = 10
        Build a forest of n_trees trees. More trees gives higher precision when querying,
        but are more expensive in terms of build time and index size.
    search_k: int, default = -1
        Query will inspect search_k nodes. A larger value will give more accurate results,
        but will take longer time.
    mmap_dir: str, default = 'auto'
        Memory-map the index to the given directory.
        This is required to make the the class pickleable.
        If None, keep everything in main memory (NON pickleable index),
        if mmap_dir is a string, it is interpreted as a directory to store the index into,
        if 'auto', create a temp dir for the index, preferably in /dev/shm on Linux.
    n_jobs: int, default = 1
        Number of parallel jobs
    verbose: int, default = 0
        Verbosity level. If verbose > 0, show tqdm progress bar on indexing and querying.

    Notes
    -----
    See more details in the annoy documentation: https://github.com/spotify/annoy#full-python-api
    """

    valid_metrics = [
        "angular",
        "euclidean",
        "manhattan",
        "hamming",
        "dot",
        "minkowski",
    ]

    def __init__(
        self,
        n_candidates: int = 5,
        metric: str = "euclidean",
        n_trees: int = 10,
        search_k: int = -1,
        mmap_dir: str = "auto",
        n_jobs: int = 1,
        verbose: int = 0,
    ):

        if annoy is None:  # pragma: no cover
            raise ImportError(
                "Please install the `annoy` package, before using this class.\n"
                "$ pip install annoy"
            ) from None

        if metric not in self.__class__.valid_metrics:
            raise ValueError(
                f"Unknown metric please try one of {self.__class__.valid_metrics}"
            )
        if metric == "minkowski":  # for compatibility
            metric = "euclidean"
        super().__init__(
            n_candidates=n_candidates,
            metric=metric,
            n_jobs=n_jobs,
        )
        self.effective_metric_ = metric
        self.verbose = verbose
        self.n_trees = n_trees
        self.search_k = search_k
        self.mmap_dir = mmap_dir
        self.index_path_source = None
        self.index_path_target = None

    def __repr__(self):
        ret_str = (
            f"{self.__class__.__name__}(n_candidates={self.n_candidates},"
            + f"effective_metric_ = {self.effective_metric_},"
            + f"verbose = {self.verbose},"
            + f"n_trees = {self.n_trees},"
            + f"search_k = {self.search_k},"
            + f"mmap_dir = {self.mmap_dir},"
            + f"n_jobs = {self.n_jobs}"
        )
        if self.index_path_source is not None:
            ret_str += (
                f" source index path={self.index_path_source} and target index"
                f" path={self.index_path_target}"
            )
        ret_str += ")"
        return ret_str

    def _fit(self, data, is_source: bool):
        if is_source:
            self._source = data
            prefix = "kiez_source"
        else:
            self._target = data
            prefix = "kiez_target"
        suffix = ".annoy"
        annoy_index = annoy.AnnoyIndex(data.shape[1], metric=self.effective_metric_)

        index_path = None
        if self.mmap_dir == "auto":
            index_path = create_tempfile_preferably_in_dir(
                prefix=prefix, suffix=suffix, directory="/dev/shm"
            )
            logging.warning(
                f"The index will be stored in {index_path}. It will NOT be deleted"
                " automatically, when this instance is destructed."
            )
        elif isinstance(self.mmap_dir, str):
            index_path = create_tempfile_preferably_in_dir(
                prefix=prefix, suffix=suffix, directory=self.mmap_dir
            )
        else:  # e.g. None
            self.mmap_dir = None

        for i, x in tqdm(
            enumerate(data),
            desc="Build Annoy",
            disable=not self.verbose,
        ):
            annoy_index.add_item(i, x.tolist())
        annoy_index.build(self.n_trees)

        if index_path is not None:
            if is_source:
                self.index_path_source = index_path
            else:
                self.index_path_target = index_path
            annoy_index.save(index_path)
            return index_path, data.shape[1]

        return annoy_index

    def _kneighbors_part(self, k, query, index, return_distance, is_self_querying):
        n_query = query.shape[0]
        query_dtype = query.dtype

        # For compatibility reasons, as each sample is considered as its own
        # neighbor, one extra neighbor will be computed.
        if is_self_querying:
            k = k + 1
            start = 1
        else:
            start = 0

        # If fewer candidates than required are found for a query,
        # we save index=-1 and distance=NaN
        neigh_ind = -np.ones((n_query, k), dtype=np.int32)
        neigh_dist = np.empty_like(neigh_ind, dtype=query_dtype) * np.nan

        annoy_index = None
        if isinstance(index, Tuple):
            err_msg = (
                "Internal error: annoy index must be either annoy.AnnoyIndex"
                + f"or Tuple[str,int], but found {index}"
            )
            try:
                index, n_features = index
                if not isinstance(index, str) or not isinstance(n_features, int):
                    raise ValueError(err_msg)
            except ValueError as exc:
                raise ValueError(err_msg) from exc
            # Load memory-mapped annoy.Index, unless it's already in main memory
            if isinstance(index, str):
                annoy_index = annoy.AnnoyIndex(
                    n_features, metric=self.effective_metric_
                )
                annoy_index.load(index)
        elif isinstance(index, annoy.AnnoyIndex):
            annoy_index = index
        assert isinstance(
            annoy_index, annoy.AnnoyIndex
        ), f"Internal error: unexpected type ({type(index)} for annoy index"

        disable_tqdm = not self.verbose
        if is_self_querying:
            n_items = annoy_index.get_n_items()

            for i in tqdm(
                range(n_items),
                desc="Query Annoy",
                disable=disable_tqdm,
            ):
                ind, dist = annoy_index.get_nns_by_item(
                    i,
                    k,
                    self.search_k,
                    include_distances=True,
                )
                ind = ind[start:]
                dist = dist[start:]
                neigh_ind[i, : len(ind)] = ind
                neigh_dist[i, : len(dist)] = dist
        else:  # if query was provided
            for i, x in tqdm(
                enumerate(query),
                desc="Query Annoy",
                disable=disable_tqdm,
            ):
                ind, dist = annoy_index.get_nns_by_vector(
                    x.tolist(),
                    k,
                    self.search_k,
                    include_distances=True,
                )
                ind = ind[start:]
                dist = dist[start:]
                neigh_ind[i, : len(ind)] = ind
                neigh_dist[i, : len(dist)] = dist

        if return_distance:
            return neigh_dist, neigh_ind
        else:
            return neigh_ind



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: BSD-3-Clause
# adapted from skhubness: https://github.com/VarIr/scikit-hubness/

from __future__ import annotations

import warnings
from typing import Tuple

import numpy as np
from sklearn.metrics import euclidean_distances
from sklearn.utils.extmath import row_norms
from sklearn.utils.validation import check_consistent_length, check_is_fitted

from .base import HubnessReduction


class DisSimLocal(HubnessReduction):
    """Hubness reduction with DisSimLocal.

    Uses the formula presented in [1]_.

    Parameters
    ----------
    k: int, default = 5
        Number of neighbors to consider for the local centroids
    squared: bool, default = True
        DisSimLocal operates on squared Euclidean distances.
        If True, return (quasi) squared Euclidean distances;
        if False, return (quasi) Eucldean distances.
    References
    ----------
    .. [1] Hara K, Suzuki I, Kobayashi K, Fukumizu K, Radovanović M (2016)
           Flattening the density gradient for eliminating spatial centrality to reduce hubness.
           In: Proceedings of the 30th AAAI conference on artificial intelligence, pp 1659–1665.
           https://www.aaai.org/ocs/index.php/AAAI/AAAI16/paper/viewPaper/12055
    """

    def __init__(self, k: int = 5, squared: bool = True, *args, **kwargs):
        super().__init__()
        self.k = k
        self.squared = squared

    def __repr__(self):
        return f"{self.__class__.__name__}(k={self.k}, squared = {self.squared})"

    def fit(
        self,
        neigh_dist: np.ndarray,
        neigh_ind: np.ndarray,
        source: np.ndarray,
        target: np.ndarray,
        assume_sorted: bool = True,
        *args,
        **kwargs,
    ) -> DisSimLocal:
        """Fit the model using target, neigh_dist, and neigh_ind as training data.

        Parameters
        ----------
        neigh_dist: np.ndarray, shape (n_samples, n_neighbors)
            Distance matrix of training objects (rows) against their
            individual k nearest neighbors (colums).
        neigh_ind: np.ndarray, shape (n_samples, n_neighbors)
            Neighbor indices corresponding to the values in neigh_dist.
        source: np.ndarray, shape (n_samples, n_features)
            source embedding, where n_samples is the number of vectors,
            and n_features their dimensionality (number of features).
        target: np.ndarray, shape (n_samples, n_features)
            Target embedding, where n_samples is the number of vectors,
            and n_features their dimensionality (number of features).
        assume_sorted: bool, default=True #noqa: DAR103
            Assume input matrices are sorted according to neigh_dist.
            If False, these are sorted here.
        *args: Ignored
            Ignored
        **kwargs: Ignored
            Ignored
        Returns
        -------
        DisSimLocal
            Fitted DisSimLocal
        Raises
        ------
        ValueError
            If self.k < 0
        TypeError
            If self.k not int
        """
        # Check equal number of rows and columns
        check_consistent_length(neigh_ind, neigh_dist)
        check_consistent_length(neigh_ind.T, neigh_dist.T)
        try:
            if self.k <= 0:
                raise ValueError(f"Expected k > 0. Got {self.k}")
        except TypeError as exc:
            raise TypeError(f"Expected k: int > 0. Got {self.k}") from exc

        k = self.k
        if k > neigh_ind.shape[1]:
            warnings.warn(
                "Neighborhood parameter k larger than provided neighbors in"
                f" neigh_dist, neigh_ind. Will reduce to k={neigh_ind.shape[1]}."
            )
            k = neigh_ind.shape[1]

        # Calculate local neighborhood centroids among the training points
        if assume_sorted:
            knn = neigh_ind[:, :k]
        else:
            mask = np.argpartition(neigh_dist, kth=k - 1)[:, :k]
            knn = np.take_along_axis(neigh_ind, mask, axis=1)
        centroids = source[knn].mean(axis=1)
        dist_to_cent = row_norms(target - centroids, squared=True)

        self.source_ = source
        self.target_ = target
        self.target_centroids_ = centroids
        self.target_dist_to_centroids_ = dist_to_cent

        return self

    def transform(
        self,
        neigh_dist: np.ndarray,
        neigh_ind: np.ndarray,
        query: np.ndarray,
        assume_sorted: bool = True,
        *args,
        **kwargs,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Transform distance between test and training data with DisSimLocal.

        Parameters
        ----------
        neigh_dist: np.ndarray, shape (n_query, n_neighbors)
            Distance matrix of test objects (rows) against their individual
            k nearest neighbors among the training data (columns).
        neigh_ind: np.ndarray, shape (n_query, n_neighbors)
            Neighbor indices corresponding to the values in neigh_dist
        query: np.ndarray, shape (n_query, n_features)
            Query entities that were used to obtain neighbors
            If none is provided use source that was provided in fit step
        assume_sorted: bool
            ignored

        Returns
        -------
        hub_reduced_dist, neigh_ind
            DisSimLocal distances, and corresponding neighbor indices
        Notes
        -----
        The returned distances are NOT sorted! If you use this class directly,
        you will need to sort the returned matrices according to hub_reduced_dist.
        """
        check_is_fitted(
            self,
            ["target_", "target_centroids_", "target_dist_to_centroids_"],
        )
        if query is None:
            query = self.source_

        n_test, n_indexed = neigh_dist.shape

        if n_indexed == 1:
            warnings.warn(
                "Cannot perform hubness reduction with a single neighbor per query. "
                "Skipping hubness reduction, and returning untransformed distances."
            )
            return neigh_dist, neigh_ind

        k = self.k
        if k > neigh_ind.shape[1]:
            warnings.warn(
                "Neighborhood parameter k larger than provided neighbors in"
                f" neigh_dist, neigh_ind. Will reduce to k={neigh_ind.shape[1]}."
            )
            k = neigh_ind.shape[1]

        # Calculate local neighborhood centroids for source objects among target objects
        mask = np.argpartition(neigh_dist, kth=k - 1)
        for i, ind in enumerate(neigh_ind):
            neigh_dist[i, :] = euclidean_distances(
                query[i].reshape(1, -1), self.target_[ind], squared=True
            )
        neigh_ind = np.take_along_axis(neigh_ind, mask, axis=1)
        knn = neigh_ind[:, :k]
        centroids = self.target_[knn].mean(axis=1)

        source_minus_centroids = query - centroids
        source_minus_centroids **= 2
        source_dist_to_centroids = source_minus_centroids.sum(axis=1)
        target_dist_to_centroids = self.target_dist_to_centroids_[neigh_ind]

        hub_reduced_dist = neigh_dist.copy()
        hub_reduced_dist -= source_dist_to_centroids[:, np.newaxis]
        hub_reduced_dist -= target_dist_to_centroids

        # DisSimLocal can yield negative dissimilarities, which can cause problems with
        # certain scikit-learn routines (e.g. in metric='precomputed' usages).
        # We, therefore, shift dissimilarities to non-negative values, if necessary.
        min_dist = hub_reduced_dist.min()
        if min_dist < 0.0:
            hub_reduced_dist += -min_dist

        # Return Euclidean or squared Euclidean distances?
        if not self.squared:
            hub_reduced_dist **= 1 / 2

        # Return the hubness reduced distances
        # These must be sorted downstream
        return hub_reduced_dist, neigh_ind



"""
Convenience methods for loading entity embeddings from knowledge graph embeddings
"""
import os
from typing import Dict, Tuple

import numpy as np


def _read_kg_ids(path) -> Dict[int, str]:
    with open(path) as in_file:
        return {
            int(line.strip().split("\t")[1]): line.strip().split("\t")[0]
            for line in in_file
        }


def _read_ent_links(path) -> Dict[str, str]:
    with open(path) as in_file:
        return {
            line.strip().split("\t")[0]: line.strip().split("\t")[1] for line in in_file
        }


def _split_emb(emb, kg_ids):
    new_emb = []
    new_ids = {}
    i = 0
    for idx, e in enumerate(emb):
        if idx in kg_ids:
            new_ids[kg_ids[idx]] = i
            new_emb.append(e)
            i += 1
    return np.array(new_emb), new_ids


def _read_openea_files(emb_dir_path, kg_path):
    emb = np.load(os.path.join(emb_dir_path, "ent_embeds.npy"))
    kg1_ids = _read_kg_ids(os.path.join(emb_dir_path, "kg1_ent_ids"))
    kg2_ids = _read_kg_ids(os.path.join(emb_dir_path, "kg2_ent_ids"))
    ent_links = _read_ent_links(os.path.join(kg_path, "ent_links"))
    return emb, kg1_ids, kg2_ids, ent_links


def _seperate_common_embedding(
    emb: np.ndarray,
    kg1_ids: Dict[int, str],
    kg2_ids: Dict[int, str],
    ent_links: Dict[str, str],
) -> Tuple[np.ndarray, np.ndarray, Dict[int, str], Dict[int, str], Dict[str, str]]:
    """Seperate single embedding array into two arrays split by knowledge graph
    Parameters
    ----------
    emb: np.ndarray
        embedding array
    kg1_ids: Dict[int,str]
        row index and entity ids of knowledge graph 1
    kg2_ids: Dict[int,str]
        row index and entity ids of knowledge graph 2
    ent_links: Dict[str,str]
        linked entities

    Returns
    -------
    emb1, emb2, kg1_ids_new, kg2_ids_new, ent_links_new
        embeddings of both knowledge graphs
        entity ids per embedding row for both knowledge graphs
        entity links
    """
    emb1, kg1_ids_new = _split_emb(emb, kg1_ids)
    emb2, kg2_ids_new = _split_emb(emb, kg2_ids)
    ent_links_new = {kg1_ids_new[e1]: kg2_ids_new[e2] for e1, e2 in ent_links.items()}
    return emb1, emb2, kg1_ids_new, kg2_ids_new, ent_links_new


def from_openea(
    emb_dir_path: str, kg_path: str
) -> Tuple[np.ndarray, np.ndarray, Dict[int, str], Dict[int, str], Dict[str, str]]:
    """Load OpenEA-type data
    Parameters
    ----------
    emb_dir_path: str
        Path to folder containing embedding
    kg_path: str
        Path to folder containing kg info

    Returns
    -------
    emb1, emb2, kg1_ids_new, kg2_ids_new, ent_links_new
        embeddings of both knowledge graphs
        entity ids per embedding row for both knowledge graphs
        entity links
    Notes
    -----
    See here for more information on the dataset structure
    https://github.com/nju-websoft/OpenEA#dataset-description
    """
    return _seperate_common_embedding(*_read_openea_files(emb_dir_path, kg_path))



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: BSD-3-Clause
from __future__ import annotations

import json
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Union

import numpy as np
from class_resolver import HintOrType

from kiez.hubness_reduction import DisSimLocal, hubness_reduction_resolver
from kiez.hubness_reduction.base import HubnessReduction, NoHubnessReduction
from kiez.neighbors import NNAlgorithm, nn_algorithm_resolver


class Kiez:
    """Performs hubness reduced nearest neighbor search for entity alignment

    Use the given algorithm to :meth:`fit` the data and calculate the
    :meth:`kneighbors`.

    Parameters
    ----------
    n_neighbors : int, default=5
        number of nearest neighbors used in search
    algorithm : :obj:`~kiez.neighbors.NNAlgorithm`, default = None
        initialised `NNAlgorithm` object that will be used for neighbor search
        If no algorithm is provided :obj:`~kiez.neighbors.Faiss` is used if available else
        :obj:`~kiez.neighbors.SklearnNN` is used with default values
    algorithm_kwargs :
        A dictionary of keyword arguments to pass to the :obj:`~kiez.neighbors.NNAlgorithm`
        if given as a class in the ``algorithm`` argument.

    hubness :
        Either an instance of a :obj:`~kiez.hubness_reduction.base.HubnessReduction`,
        the class for a :obj:`~kiez.hubness_reduction.base.HubnessReduction` that should
        be instantiated, the name of the hubness reduction method, or if None, defaults
        to no hubness reduction.
    hubness_kwargs :
        A dictionary of keyword arguments to pass to the :obj:`~kiez.hubness_reduction.base.HubnessReduction`
        if given as a class in the ``hubness`` argument.

    Examples
    --------
    >>> from kiez import Kiez
    >>> import numpy as np
    >>> # create example data
    >>> rng = np.random.RandomState(0)
    >>> source = rng.rand(100,50)
    >>> target = rng.rand(100,50)
    >>> # fit and get neighbors
    >>> k_inst = Kiez()
    >>> k_inst.fit(source, target)
    >>> nn_dist, nn_ind = k_inst.kneighbors()

    Using a specific algorithm and hubness reduction

    >>> from kiez import Kiez
    >>> import numpy as np
    >>> # create example data
    >>> rng = np.random.RandomState(0)
    >>> source = rng.rand(100,50)
    >>> target = rng.rand(100,50)
    >>> # prepare algorithm and hubness reduction
    >>> from kiez.neighbors import NMSLIB
    >>> hnsw = NMSLIB(n_candidates=10)
    >>> from kiez.hubness_reduction import CSLS
    >>> hr = CSLS()
    >>> # fit and get neighbors
    >>> k_inst = Kiez(n_neighbors=5, algorithm=hnsw, hubness=hr)
    >>> k_inst.fit(source, target)
    >>> nn_dist, nn_ind = k_inst.kneighbors()

    You can also initalize Kiez via a json file

    >>> kiez = Kiez.from_path("tests/example_conf.json")
    """

    def __init__(
        self,
        n_neighbors: int = 5,
        algorithm: HintOrType[NNAlgorithm] = None,
        algorithm_kwargs: Optional[Dict[str, Any]] = None,
        hubness: HintOrType[HubnessReduction] = None,
        hubness_kwargs: Optional[Dict[str, Any]] = None,
    ):
        if not np.issubdtype(type(n_neighbors), np.integer):
            raise TypeError(
                f"n_neighbors does not take {type(n_neighbors)} value, enter"
                " integer value"
            )
        elif n_neighbors <= 0:
            raise ValueError(f"Expected n_neighbors > 0. Got {n_neighbors}")
        self.n_neighbors = n_neighbors
        if algorithm is None and algorithm_kwargs is None:
            algorithm_kwargs = {"n_candidates": n_neighbors}
        if algorithm is None:
            try:
                self.algorithm = nn_algorithm_resolver.make("Faiss", algorithm_kwargs)
            except ImportError:
                self.algorithm = nn_algorithm_resolver.make(
                    "SklearnNN", algorithm_kwargs
                )
        else:
            self.algorithm = nn_algorithm_resolver.make(algorithm, algorithm_kwargs)
        assert self.algorithm
        self.hubness = hubness_reduction_resolver.make(hubness, hubness_kwargs)
        self._check_algorithm_hubness_compatibility()

    def __repr__(self):
        return (
            f"Kiez(n_neighbors: {self.n_neighbors}, algorithm: {self.algorithm},"
            f" hubness: {self.hubness})"
            f" {self.algorithm._describe_source_target_fitted()}"
        )

    @classmethod
    def from_path(cls, path: Union[str, Path]) -> Kiez:
        """Load a Kiez instance from configuration in a JSON file, based on its path."""
        with open(path) as file:
            return cls(**json.load(file))

    def _kcandidates(
        self, query_points, *, s_to_t=True, k=None, return_distance=True
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        if k is None:
            k = self.algorithm.n_candidates

        # The number of candidates must not be less than the number of neighbors used downstream
        if k < self.n_neighbors:
            k = self.n_neighbors
        return self.algorithm.kneighbors(
            k=k,
            query=query_points,
            s_to_t=s_to_t,
            return_distance=return_distance,
        )

    def _check_algorithm_hubness_compatibility(self):
        if isinstance(self.hubness, DisSimLocal):
            if self.algorithm.metric in ["euclidean", "minkowski"]:
                self.hubness.squared = False
                if hasattr(self.algorithm, "p"):
                    if self.algorithm.p != 2:
                        raise ValueError(
                            "DisSimLocal only supports squared Euclidean distances. If"
                            " the provided NNAlgorithm has a `p` parameter it must be"
                            f" set to p=2. Now it is p={self.algorithm.p}"
                        )
            elif self.algorithm.metric in ["sqeuclidean"]:
                self.hubness.squared = True
            else:
                raise ValueError(
                    "DisSimLocal only supports squared Euclidean distances, not"
                    f" metric={self.algorithm.metric}."
                )

    def fit(self, source, target=None) -> Kiez:
        """Fits the algorithm and hubness reduction method

        Parameters
        ----------
        source : matrix of shape (n_samples, n_features)
            embeddings of source entities
        target : matrix of shape (m_samples, n_features)
            embeddings of target entities. If none given, uses the source.

        Returns
        -------
        Kiez
            Fitted kiez instance
        """
        self.algorithm.fit(source, target)
        if target is None:
            target = source
        if not isinstance(self.hubness, NoHubnessReduction):
            neigh_dist_t_to_s, neigh_ind_t_to_s = self._kcandidates(
                target,
                s_to_t=False,
                k=self.algorithm.n_candidates,
                return_distance=True,
            )
            self.hubness.fit(
                neigh_dist_t_to_s,
                neigh_ind_t_to_s,
                source,
                target,
                assume_sorted=False,
            )
        return self

    def kneighbors(
        self,
        source_query_points=None,
        k=None,
        return_distance=True,
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Retrieves the k-nearest neighbors using the supplied nearest neighbor algorithm and hubness reduction method.

        Parameters
        ----------
        source_query_points : matrix of shape (n_samples, n_features), default = None
            subset of source entity embeddings
            if `None` all source entities are used for querying
        k : int, default = None
            number of nearest neighbors to search for
        return_distance : bool, default = True
            Whether to return distances
            If `False` only indices are returned

        Returns
        -------
        neigh_dist : ndarray of shape (n_queries, n_neighbors)
            Array representing the distance between source and target entities
            only present if return_distance=True.
        neigh_ind : ndarray of shape (n_queries, n_neighbors)
            Indices of the nearest points in the population matrix.
        """
        # function loosely adapted from skhubness: https://github.com/VarIr/scikit-hubness

        if k is None:
            n_neighbors = self.n_neighbors
        else:
            n_neighbors = k
        # First obtain candidate neighbors
        query_dist, query_ind = self._kcandidates(
            source_query_points, return_distance=True
        )
        query_dist = np.atleast_2d(query_dist)
        query_ind = np.atleast_2d(query_ind)

        # Second, reduce hubness
        hubness_reduced_query_dist, query_ind = self.hubness.transform(
            query_dist,
            query_ind,
            source_query_points,
            assume_sorted=True,
        )
        # Third, sort hubness reduced candidate neighbors to get the final k neighbors
        kth = np.arange(n_neighbors)
        mask = np.argpartition(hubness_reduced_query_dist, kth=kth)[:, :n_neighbors]
        hubness_reduced_query_dist = np.take_along_axis(
            hubness_reduced_query_dist, mask, axis=1
        )
        query_ind = np.take_along_axis(query_ind, mask, axis=1)

        if return_distance:
            result = hubness_reduced_query_dist, query_ind
        else:
            result = query_ind
        return result



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: BSD-3-Clause
# Author: Roman Feldbauer (version for scikit-hubness)
#         Daniel Obraczka (adaptation for kiez)
# PEP 563: Postponed Evaluation of Annotations

from __future__ import annotations

import numpy as np

from kiez.neighbors.neighbor_algorithm_base import NNAlgorithm

try:
    import nmslib
except ImportError:  # pragma: no cover
    nmslib = None


class NMSLIB(NNAlgorithm):
    """
    Wrapper for hierarchical navigable small world graphs based approximate nearest neighbor search implementation from NMSLIB

    Parameters
    ----------
    n_candidates: int, default = 5
        number of nearest neighbors used in search
    metric: str, default = 'euclidean'
        distance measure used in search
        possible measures are found in :obj:`NMSLIB.valid_metrics`
    method: str, default = 'hnsw',
        ANN method to use. Currently, only 'hnsw' is supported.
    M: int, default = 16
        maximum number of neighbors in zero or above layers of hierarchical graph
    post_processing: int, default = 2
        More post processing means longer index creation,
        and higher retrieval accuracy.
    ef_construction: int, default = 200
        higher value improves quality of constructed graph but leads to longer indexing times
    n_jobs: int, default = 1
        Number of parallel jobs
    verbose: int, default = 0
        Verbosity level. If verbose > 0, show tqdm progress bar on indexing and querying.

    Notes
    -----
    See the nmslib documentation for more details: https://github.com/nmslib/nmslib/blob/master/manual/methods.md
    """

    valid_metrics = [
        "euclidean",
        "l2",
        "minkowski",
        "squared_euclidean",
        "sqeuclidean",
        "cosine",
        "cosinesimil",
    ]

    def __init__(
        self,
        n_candidates: int = 5,
        metric: str = "euclidean",
        method: str = "hnsw",
        M: int = 16,  # noqa: N803
        post_processing: int = 2,
        ef_construction: int = 200,
        n_jobs: int = 1,
        verbose: int = 0,
    ):

        self.space = None
        if nmslib is None:  # pragma: no cover
            raise ImportError(
                "Please install the `nmslib` package, before using this class.\n"
                "$ pip install nmslib"
            )

        if metric in [
            "euclidean",
            "l2",
            "minkowski",
            "squared_euclidean",
            "sqeuclidean",
        ]:
            if metric in ["squared_euclidean", "sqeuclidean"]:
                metric = "sqeuclidean"
            else:
                metric = "euclidean"
            self.space = "l2"
        elif metric in ["cosine", "cosinesimil"]:
            self.space = "cosinesimil"
        elif metric not in self.__class__.valid_metrics:
            raise ValueError(
                f"Unknown metric please try one of {self.__class__.valid_metrics}"
            )
        assert self.space in [
            "l2",
            "cosinesimil",
        ], f"Internal: self.space={self.space} not allowed"
        super().__init__(n_candidates=n_candidates, metric=metric, n_jobs=n_jobs)
        self.verbose = verbose
        self.method = method
        self.M = M  # noqa: N803
        self.post_processing = post_processing
        self.ef_construction = ef_construction

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(n_candidates={self.n_candidates}, "
            + f"verbose = {self.verbose}, "
            + f"method = {self.method}, "
            + f"M = {self.M}, "
            + f"space = {self.space}, "
            + f"post_processing = {self.post_processing}, "
            + f"ef_construction = {self.ef_construction}, "
            + f"n_jobs = {self.n_jobs}) "
        )

    def _fit(self, data, is_source: bool):
        method = self.method
        post_processing = self.post_processing
        big_m = self.M
        ef_construction = self.ef_construction

        hnsw_index = nmslib.init(method=method, space=self.space)
        hnsw_index.addDataPointBatch(data)
        hnsw_index.createIndex(
            {
                "M": big_m,
                "efConstruction": ef_construction,
                "post": post_processing,
                "indexThreadQty": self.n_jobs,
            },
            print_progress=(self.verbose >= 2),
        )
        return hnsw_index

    def _kneighbors(self, k, query, index, return_distance, is_self_querying):
        # Fetch the neighbor candidates
        neigh_ind_dist = index.knnQueryBatch(query, k=k, num_threads=self.n_jobs)

        # If fewer candidates than required are found for a query,
        # we save index=-1 and distance=NaN
        n_query = query.shape[0]
        neigh_ind = -np.ones((n_query, k), dtype=np.int32)
        neigh_dist = np.empty_like(neigh_ind, dtype=query.dtype) * np.nan

        for i, (ind, dist) in enumerate(neigh_ind_dist):
            neigh_ind[i, : ind.size] = ind
            neigh_dist[i, : dist.size] = dist

        # Convert cosine similarities to cosine distances
        if self.space == "cosinesimil":
            neigh_dist *= -1
            neigh_dist += 1
        elif self.space == "l2":
            if self.metric == "sqeuclidean":
                neigh_dist **= 2
            elif self.metric == "euclidean":
                neigh_dist = np.sqrt(neigh_dist)

        if return_distance:
            return neigh_dist, neigh_ind
        else:
            return neigh_ind



# documentation copied in parts from scikit-learn
from sklearn.neighbors import VALID_METRICS, NearestNeighbors

from kiez.neighbors.neighbor_algorithm_base import NNAlgorithm


class SklearnNN(NNAlgorithm):
    """Wrapper for scikit learn's NearestNeighbors class

    Parameters
    ----------
    n_candidates: int
        number of nearest neighbors used in search
    algorithm : {'auto', 'ball_tree', 'kd_tree', 'brute'}, default='auto'
        Algorithm used to compute the nearest neighbors:

        - 'ball_tree' will use :class:`sklearn.neighbors.BallTree`
        - 'kd_tree' will use :class:`sklearn.neighbors.KDTree`
        - 'brute' will use a brute-force search.
        - 'auto' will attempt to decide the most appropriate algorithm
          based on the values passed to :meth:`~kiez.neighbors.neighbor_algorithm_base.NNAlgorithm.fit` method.
    leaf_size : int, default=30
        Leaf size passed to BallTree or KDTree.  This can affect the
        speed of the construction and query, as well as the memory
        required to store the tree.  The optimal value depends on the
        nature of the problem.
    metric: str, default = 'minkowski'
        distance measure used in search
        default is minkowski with p=2, which is equivlanet to euclidean
        possible measures are found in :obj:`SklearnNN.valid_metrics`
    p: int, default=2
        Parameter for the Minkowski metric.
        When p = 1, this is equivalent to using manhattan_distance (l1),
        and euclidean_distance (l2) for p = 2.
        For arbitrary p, minkowski_distance (l_p) is used.
    metric_params: dict, default=None
        Additional keyword arguments for the metric function.
        metric_params
    n_jobs : int, default=None
        The number of parallel jobs to run for neighbors search.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.
    Notes
    -----
    See also scikit learn's guide: https://scikit-learn.org/stable/modules/neighbors.html#unsupervised-neighbors
    """

    valid_metrics = VALID_METRICS

    def __init__(
        self,
        n_candidates=5,
        algorithm="auto",
        leaf_size=30,
        metric="minkowski",
        p=2,
        metric_params=None,
        n_jobs=None,
    ):
        super().__init__(n_candidates=n_candidates, metric=metric, n_jobs=n_jobs)
        self.algorithm = algorithm
        self.leaf_size = leaf_size
        self.p = p
        self.metric_params = metric_params

    def __repr__(self):
        ret_str = (
            f"{self.__class__.__name__}(n_candidates={self.n_candidates},"
            + f"algorithm={self.algorithm},"
            + f"leaf_size={self.leaf_size},"
            + f"metric={self.metric},"
            + f"n_jobs={self.n_jobs} "
        )
        if (
            hasattr(self, "source_index")
            and self.source_index._fit_method != self.algorithm
        ):
            return ret_str + f" and effective algo is {self.source_index._fit_method}"
        ret_str += ")"
        return ret_str

    def _fit(self, data, is_source: bool):
        nn = NearestNeighbors(
            n_neighbors=self.n_candidates,
            algorithm=self.algorithm,
            leaf_size=self.leaf_size,
            metric=self.metric,
            p=self.p,
            metric_params=self.metric_params,
            n_jobs=self.n_jobs,
        )
        nn.fit(data)
        return nn

    def _kneighbors(self, k, query, index, return_distance, is_self_querying):
        if is_self_querying:
            return index.kneighbors(
                X=None, n_neighbors=k, return_distance=return_distance
            )
        return index.kneighbors(X=query, n_neighbors=k, return_distance=return_distance)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: BSD-3-Clause
# Author: Roman Feldbauer

import logging
from tempfile import NamedTemporaryFile, mkstemp
from typing import Optional

__all__ = ["create_tempfile_preferably_in_dir"]


def create_tempfile_preferably_in_dir(
    suffix: Optional[str] = None,
    prefix: Optional[str] = None,
    directory: Optional[str] = None,
    persistent: Optional[bool] = False,
):
    """Create a temporary file with precedence for directory if possible, in TMP otherwise.
    For example, this is useful to try to save into /dev/shm.

    Parameters
    ---------
    suffix: str
        suffix of tempfile
    prefix: str
        prefix of tempfile
    directory: str
        directory where tempfile should preferably be created
    persistent: bool
        If True create a persistent file

    Returns
    -------
    path
        string path of tempfile
    """
    temp_file = mkstemp if persistent else NamedTemporaryFile
    try:
        handle = temp_file(suffix=suffix, prefix=prefix, dir=directory)  # type: ignore
        warn = False
    except FileNotFoundError:
        handle = temp_file(suffix=suffix, prefix=prefix, dir=None)  # type: ignore
        warn = True

    # Extract the path (as string)
    try:
        path = handle.name
    except AttributeError:
        _, path = handle

    if warn:
        logging.warning(
            f"Could not create temp file in {directory}. Instead, the path is {path}."
        )
    return path



import logging
import warnings
from typing import Optional

import numpy as np

from kiez.neighbors.neighbor_algorithm_base import NNAlgorithm

try:
    import faiss
except ImportError:  # pragma: no cover
    faiss = None
try:
    import autofaiss
except ImportError:  # pragma: no cover
    autofaiss = None


class Faiss(NNAlgorithm):
    """Wrapper for `faiss` library.

    Faiss implements a number of (A)NN algorithms and enables the use of GPUs.
    If it is installed and you let it, kiez utilizes the `autofaiss` package to
    find the appropriate indexing structure and optimizes the hyperparameters of
    the algorithm

    Parameters
    ----------
    n_candidates: int, default = 5
        number of nearest neighbors used in search
    metric: str, default = 'l2'
        distance measure used in search
        possible measures are found in :obj:`Faiss.valid_metrics`
        Euclidean is the same as l2, expect for taking the sqrt of the result
    index_key: str, default = None
        index name to use
        If none is provided will determine the best automatically
        Else will use it as input for :meth:`faiss.index_factory`
    index_param: str, default = None
        Hyperparameter string for the indexing algorithm
        See https://github.com/facebookresearch/faiss/wiki/Index-IO,-cloning-and-hyper-parameter-tuning#auto-tuning-the-runtime-parameters for info
    use_gpu: bool
        If true uses all available gpus

    Examples
    --------
    >>> import numpy as np
    >>> from kiez import Kiez
    >>> source = np.random.rand(1000, 512)
    >>> target = np.random.rand(100, 512)
    >>> k_inst = Kiez(algorithm="Faiss")
    >>> k_inst.fit(source, target)

    get info about selected indices

    >>> k_inst.algorithm.source_index_infos["index_key"]
    'HNSW15'

    >>> k_inst = Kiez(algorithm="Faiss",algorithm_kwargs={"metric":"euclidean","index_key":"Flat"})


    supply hyperparameters for indexing algorithm

    >>> k_inst = Kiez(algorithm="Faiss",algorithm_kwargs={"index_key":"HNSW32","index_param":"efSearch=16383"})

    Notes
    -----
    For details about configuring faiss consult their wiki: https://github.com/facebookresearch/faiss/wiki
    For details about autofaiss see their documentation: https://criteo.github.io/autofaiss/
    """

    valid_metrics = ["l2", "euclidean"]
    valid_spaces = ["l2"]

    def __init__(
        self,
        n_candidates: int = 5,
        metric: str = "l2",
        index_key: Optional[str] = None,
        index_param: Optional[str] = None,
        use_gpu: bool = False,
        verbose: int = logging.WARNING,
    ):
        if faiss is None:  # pragma: no cover
            raise ImportError(
                "Please install the `faiss` package, before using this class.\nSee here"
                " for installation instructions:"
                " https://github.com/facebookresearch/faiss/blob/main/INSTALL.md"
            )
        if metric not in self.__class__.valid_metrics:
            raise ValueError(
                f"Unknown metric {metric}, please use one of {self.valid_metrics}"
            )
        if metric not in self.__class__.valid_spaces:
            if metric == "euclidean":
                self.space = "l2"
        else:
            self.space = metric
        super().__init__(n_candidates=n_candidates, metric=metric, n_jobs=None)
        use_auto_tune = autofaiss is not None
        # check index string
        if index_key:
            try:
                faiss.index_factory(1, index_key)
            except RuntimeError as exc:
                raise ValueError(
                    f'Could not parse index "{index_key}".\n Please consult the faiss'
                    " wiki to create a correct instruction:"
                    " https://github.com/facebookresearch/faiss/wiki/The-index-factory"
                ) from exc
            # user seems to know what they want so no tuning
            if index_param or index_key == "Flat":
                use_auto_tune = False
        elif index_param:
            warnings.warn(
                "Index key not set but hyperparameter given. Are you sure this is"
                " intended?"
            )
        else:
            # no index and no hyperparams so check
            # if autofaiss is available
            if autofaiss is None:  # pragma: no cover
                warnings.warn(
                    "Please install the `autofaiss` package, to enable automatic index"
                    " selection.\nYou can install `autofaiss` via: pip install"
                    " autofaiss\n Will use `Flat` index for now, but there are probably"
                    " better choices..."
                )
                use_auto_tune = False
        self.index_key = index_key
        self.index_param = index_param
        self.use_auto_tune = use_auto_tune
        self.use_gpu = use_gpu
        self.index_infos = None
        self.verbose = verbose

    def _source_target_repr(self, is_source: bool):
        ret_str = f"{self.__class__.__name__}(n_candidates={self.n_candidates},metric={self.metric},"
        if is_source:
            ret_str += (
                f"index_key={self.source_index_key},"
                f" index_param={{{self.source_index_param}}},"
            )
        else:
            ret_str += (
                f"index_key={self.target_index_key},"
                f" index_param={{{self.target_index_param}}},"
            )
        ret_str += f"use_auto_tune={self.use_auto_tune}, use_gpu={self.use_gpu})"
        return ret_str

    def __repr__(self):
        if hasattr(self, "source_index_key") and hasattr(self, "target_index_key"):
            ret_str = (
                f"Source: {self._source_target_repr(True)}, "
                f"Target: {self._source_target_repr(False)}"
            )
        elif hasattr(self, "source_index_key"):
            ret_str = f"{self._source_target_repr(True)}"
        else:
            ret_str = (
                f"{self.__class__.__name__}(n_candidates={self.n_candidates},"
                + f"metric={self.metric},"
                + f"index_key={self.index_key},"
                + f"index_param={{{self.index_param}}},"
                + f"use_auto_tune={self.use_auto_tune},"
                + f"use_gpu={self.use_gpu})"
            )
        return ret_str

    def _to_float32(self, data):
        if not data.dtype == "float32":
            return data.astype("float32")
        return data

    def _fit(self, data, is_source: bool):
        dim = data.shape[1]
        if self.use_auto_tune:
            index, index_infos = autofaiss.build_index(
                self._to_float32(data),
                index_key=self.index_key,
                index_param=self.index_param,
                metric_type=self.space,
                save_on_disk=False,
                use_gpu=self.use_gpu,
                verbose=self.verbose,
            )
            if is_source:
                self.source_index_key = index_infos["index_key"]
                self.source_index_param = index_infos["index_param"]
                self.source_index_infos = index_infos
            else:
                self.target_index_key = index_infos["index_key"]
                self.target_index_param = index_infos["index_param"]
                self.target_index_infos = index_infos
        else:
            index = faiss.index_factory(dim, self.index_key)
            params = faiss.ParameterSpace()
            if self.use_gpu:
                index = faiss.index_cpu_to_all_gpus(index)
                params = faiss.GpuParameterSpace()
            if self.index_param is not None:
                params.set_index_parameters(index, self.index_param)
            index.add(self._to_float32(data))
        return index

    def _kneighbors(self, k, query, index, return_distance, is_self_querying):
        if is_self_querying:
            dist, ind = index.search(self._to_float32(self.source_), k)
        else:
            dist, ind = index.search(self._to_float32(query), k)
        if return_distance:
            if self.metric == "euclidean":
                dist = np.sqrt(dist)
            return dist, ind
        else:
            return ind



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: BSD-3-Clause
# adapted from skhubness: https://github.com/VarIr/scikit-hubness/

from class_resolver import ClassResolver

from .base import HubnessReduction, NoHubnessReduction
from .csls import CSLS
from .dis_sim import DisSimLocal
from .local_scaling import LocalScaling
from .mutual_proximity import MutualProximity

hubness_reduction_resolver = ClassResolver.from_subclasses(
    base=HubnessReduction,
    default=NoHubnessReduction,
)

#: Supported hubness reduction algorithms
__all__ = [
    "hubness_reduction_resolver",
    "NoHubnessReduction",
    "LocalScaling",
    "MutualProximity",
    "DisSimLocal",
    "CSLS",
]



from typing import List, Type

from kiez.neighbors import NNAlgorithm, nn_algorithm_resolver


def available_ann_algorithms() -> List[Type[NNAlgorithm]]:
    """Get available approximate nearest neighbor algorithms
    Returns
    -------
    algorithms: List[Type[NNAlgorithm]]
        A tuple of available algorithms
    """
    possible = ["NMSLIB", "NNG", "Annoy", "Faiss"]
    available = []
    for ann in possible:
        try:
            nn_algorithm_resolver.make(ann)
            available.append(nn_algorithm_resolver.lookup(ann))
        except ImportError:
            pass
    return available



"""
Calculate evaluation metrics such as hits@k
"""
from typing import Any, Dict, List, Union

import numpy as np


def _hits_from_ndarray(nn_ind, gold, k, hits_counter):
    for hits_at_k in k:
        for i, _ in enumerate(nn_ind):
            if i in gold and gold[i] in nn_ind[i][:hits_at_k]:
                hits_counter[hits_at_k] += 1
    return hits_counter


def _hits_from_dict(nn_ind, gold, k, hits_counter):
    for hits_at_k in k:
        for i, v in nn_ind.items():
            if i in gold and gold[i] in v[:hits_at_k]:
                hits_counter[hits_at_k] += 1
    return hits_counter


def hits(
    nn_ind: Union[np.ndarray, Dict[Any, List]],
    gold: Dict[Any, Any],  # source -> target
    k=None,
) -> Dict[int, float]:
    """Show hits@k

    Parameters
    ----------
    nn_ind:  array or dict
        Contains the indices of the nearest neighbors for source entities.
    gold: dict
        Map of source indices to the respective target indices
    k: array
        Which ks should be evaluated

    Returns
    -------
    hits: dict
        k: relative hits@k values

    Examples
    --------
    >>> from kiez.evaluate import hits
    >>> import numpy as np
    >>> nn_ind = np.array([[1, 2, 3], [2, 3, 4], [3, 4, 5], [4, 5, 6]])
    >>> gold = {0: 2, 1: 4, 2: 3, 3: 4}
    >>> hits(nn_ind, gold)
    {1: 0.5, 5: 1.0, 10: 1.0}
    """
    if k is None:
        k = [1, 5, 10]
    k.sort()
    hits_counter = {hits_at_k: 0 for hits_at_k in k}
    if isinstance(nn_ind, (np.ndarray, list)):
        hits_counter = _hits_from_ndarray(nn_ind, gold, k, hits_counter)
    elif isinstance(nn_ind, Dict):
        hits_counter = _hits_from_dict(nn_ind, gold, k, hits_counter)
    return {hits_at_k: k_val / len(gold) for hits_at_k, k_val in hits_counter.items()}



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: BSD-3-Clause
# Author: Roman Feldbauer (adaptions for scikit-hubness)
#         Daniel Obraczka (adaptions for kiez)
# PEP 563: Postponed Evaluation of Annotations

from __future__ import annotations

import logging

import numpy as np
from tqdm.auto import tqdm

from kiez.io.temp_file_handling import create_tempfile_preferably_in_dir
from kiez.neighbors.neighbor_algorithm_base import NNAlgorithmWithJoblib

try:
    import ngtpy  # noqa: autoimport

except ImportError:  # pragma: no cover
    ngtpy = None


__all__ = [
    "NNG",
]


class NNG(NNAlgorithmWithJoblib):
    """
    Wrapper for NGT's graph based approximate nearest neighbor search

    Parameters
    ----------
    n_candidates: int
        number of nearest neighbors used in search
    metric: str, default = 'euclidean'
        distance measure used in search
        possible measures are found in :obj:`NNG.valid_metrics`
    index_dir: str default = 'auto'
        Store the index in the given directory.
        If None, keep the index in main memory (NON pickleable index),
        If index_dir is a string, it is interpreted as a directory to store the index into,
        if 'auto', create a temp dir for the index, preferably in /dev/shm on Linux.
        Note: The directory/the index will NOT be deleted automatically.
    edge_size_for_creation: int, default = 80
        Increasing ANNG edge size improves retrieval accuracy at the cost of more time
    edge_size_for_search: int, default = 40
        Increasing ANNG edge size improves retrieval accuracy at the cost of more time
    epsilon: float, default 0.1
        Trade-off in ANNG between higher accuracy (larger epsilon) and shorter query time (smaller epsilon)
    n_jobs: int, default = 1
        Number of parallel jobs
    verbose: int, default = 0
        Verbosity level. If verbose > 0, show tqdm progress bar on indexing and querying.

    Notes
    -----
    See the NGT documentation for more details: https://github.com/yahoojapan/NGT/blob/master/python/README-ngtpy.md

    NNG stores the index to a directory specified in `index_dir`.
    The index is persistent, and will NOT be deleted automatically.
    It is the user's responsibility to take care of deletion,
    when required.
    """

    valid_metrics = [
        "manhattan",
        "L1",
        "euclidean",
        "L2",
        "minkowski",
        "sqeuclidean",
        "Angle",
        "Normalized Angle",
        "Cosine",
        "Normalized Cosine",
        "Hamming",
        "Jaccard",
    ]
    _internal_distance_type = {
        "manhattan": "L1",
        "euclidean": "L2",
        "minkowski": "L2",
        "sqeuclidean": "L2",
    }

    def __init__(
        self,
        n_candidates: int = 5,
        metric: str = "euclidean",
        index_dir: str = "auto",
        edge_size_for_creation: int = 80,
        edge_size_for_search: int = 40,
        epsilon: float = 0.1,
        n_jobs: int = 1,
        verbose: int = 0,
    ):

        if ngtpy is None:  # pragma: no cover
            raise ImportError(
                "Please install the `ngt` package, before using this class.\n"
                "$ pip3 install ngt"
            ) from None
        super().__init__(n_candidates=n_candidates, metric=metric, n_jobs=n_jobs)
        # Map common distance names to names used by ngt
        try:
            self.effective_metric_ = NNG._internal_distance_type[self.metric]
        except KeyError:
            self.effective_metric_ = self.metric
        if self.effective_metric_ not in NNG.valid_metrics:
            raise ValueError(
                f"Unknown distance/similarity measure: {self.effective_metric_}. "
                f"Please use one of: {NNG.valid_metrics}."
            )
        self.verbose = verbose
        self.index_dir = index_dir
        self._index_dir_plausibility_check()
        self.edge_size_for_creation = edge_size_for_creation
        self.edge_size_for_search = edge_size_for_search
        self.epsilon = epsilon
        self.index_path_source = None
        self.index_path_target = None

    def __repr__(self):
        ret_str = (
            f"{self.__class__.__name__}(n_candidates={self.n_candidates},"
            + f"index_dir = {self.index_dir},"
            + f"edge_size_for_creation = {self.edge_size_for_creation},"
            + f"edge_size_for_search = {self.edge_size_for_search},"
            + f"epsilon = {self.epsilon},"
            + f"n_jobs = {self.n_jobs},"
            + f"verbose = {self.verbose}"
        )
        if self.index_path_source is not None:
            ret_str += (
                f" source index path={self.index_path_source} and target index"
                f" path={self.index_path_target}"
            )
        if self.metric != self.effective_metric_:
            return ret_str + f" and effective algo is {self.effective_metric_}"
        ret_str += ")"
        return ret_str

    def _index_dir_plausibility_check(self):
        if not (self.index_dir is None or isinstance(self.index_dir, str)):
            raise TypeError(
                "NNG requires to write an index to the filesystem. "
                "Please provide a valid path with parameter `index_dir`."
            )

    def _fit(self, data, is_source: bool):
        if is_source:
            prefix = "kiez_source"
        else:
            prefix = "kiez_target"

        index_path = None
        # Set up a directory to save the index to
        suffix = ".anng"
        if self.index_dir in ["auto"]:
            index_path = create_tempfile_preferably_in_dir(
                prefix=prefix, suffix=suffix, directory="/dev/shm"
            )
            logging.warning(
                f"The index will be stored in {index_path}. It will NOT be deleted"
                " automatically, when this instance is destructed."
            )
        elif isinstance(self.index_dir, str):
            index_path = create_tempfile_preferably_in_dir(
                prefix=prefix, suffix=suffix, directory=self.index_dir
            )
        elif self.index_dir is None:
            index_path = create_tempfile_preferably_in_dir(prefix=prefix, suffix=suffix)

        # Create the ANNG index, insert data
        ngtpy.create(
            path=index_path,
            dimension=data.shape[1],
            edge_size_for_creation=self.edge_size_for_creation,
            edge_size_for_search=self.edge_size_for_search,
            distance_type=self.effective_metric_,
        )
        index_obj = ngtpy.Index(index_path)
        index_obj.batch_insert(data, num_threads=self.n_jobs)
        index_obj.save()

        if index_path is not None:
            if is_source:
                self.index_path_source = index_path
            else:
                self.index_path_target = index_path

        # Keep index in memory or store in path
        if self.index_dir is None:
            return index_obj
        return index_path

    def _kneighbors_part(self, k, query, index, return_distance, is_self_querying):
        index = (
            ngtpy.Index(index) if isinstance(index, str) else index  # load if is path
        )
        n_query = query.shape[0]
        query_dtype = query.dtype

        # For compatibility reasons, as each sample is considered as its own
        # neighbor, one extra neighbor will be computed.
        if is_self_querying:
            k = k + 1
            start = 1
        else:
            start = 0

        # If fewer candidates than required are found for a query,
        # we save index=-1 and distance=NaN
        neigh_ind = -np.ones((n_query, k), dtype=np.int32)
        if return_distance:
            neigh_dist = np.empty_like(neigh_ind, dtype=query_dtype) * np.nan

        disable_tqdm = not self.verbose
        if is_self_querying:
            for i in tqdm(
                range(n_query),
                desc="Query NNG",
                disable=disable_tqdm,
            ):
                query = index.get_object(i)
                response = index.search(
                    query=query,
                    size=k,
                    with_distance=return_distance,
                    epsilon=self.epsilon,
                )
                if return_distance:
                    ind, dist = [np.array(arr) for arr in zip(*response)]
                else:
                    ind = response
                ind = ind[start:]
                neigh_ind[i, : len(ind)] = ind
                if return_distance:
                    dist = dist[start:]
                    neigh_dist[i, : len(dist)] = dist
        else:  # if query was provided
            for i, x in tqdm(
                enumerate(query),
                desc="Query NNG",
                disable=disable_tqdm,
            ):
                response = index.search(
                    query=x,
                    size=k,
                    with_distance=return_distance,
                    epsilon=self.epsilon,
                )
                if return_distance:
                    ind, dist = [np.array(arr) for arr in zip(*response)]
                else:
                    ind = response
                ind = ind[start:]
                neigh_ind[i, : len(ind)] = ind
                if return_distance:
                    dist = dist[start:]
                    neigh_dist[i, : len(dist)] = dist

        if return_distance and self.metric == "sqeuclidean":
            neigh_dist **= 2

        if return_distance:
            return neigh_dist, neigh_ind
        else:
            return neigh_ind



from class_resolver import ClassResolver

from kiez.neighbors.approximate.faiss import Faiss
from kiez.neighbors.approximate.nmslib import NMSLIB
from kiez.neighbors.approximate.nng import NNG
from kiez.neighbors.approximate.random_projection_trees import Annoy
from kiez.neighbors.exact.sklearn_nearest_neighbors import SklearnNN
from kiez.neighbors.neighbor_algorithm_base import NNAlgorithm, NNAlgorithmWithJoblib

__all__ = [
    "nn_algorithm_resolver",
    "NMSLIB",
    "NNG",
    "Annoy",
    "SklearnNN",
    "Faiss",
    "NNAlgorithm",
]

nn_algorithm_resolver = ClassResolver.from_subclasses(
    base=NNAlgorithm,
    skip={  # Skip being able to resolve intermediate base classes
        NNAlgorithmWithJoblib,
    },
    default=SklearnNN,
)



# -*- coding: utf-8 -*-
# adapted from skhubness
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC, abstractmethod


class HubnessReduction(ABC):
    """Base class for hubness reduction."""

    @abstractmethod
    def __init__(self, **kwargs):
        pass

    @abstractmethod
    def fit(
        self, neigh_dist, neigh_ind, source, target, assume_sorted, *args, **kwargs
    ):
        pass  # pragma: no cover

    @abstractmethod
    def transform(self, neigh_dist, neigh_ind, query, assume_sorted, *args, **kwargs):
        pass  # pragma: no cover


class NoHubnessReduction(HubnessReduction):
    """Compatibility class for neighbor search without hubness reduction."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def fit(self, *args, **kwargs):
        pass  # pragma: no cover

    def __repr__(self):
        return "NoHubnessReduction"

    def transform(
        self, neigh_dist, neigh_ind, query, assume_sorted=True, *args, **kwargs
    ):
        return neigh_dist, neigh_ind



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: BSD-3-Clause
# adapted from skhubness: https://github.com/VarIr/scikit-hubness/
"""
Estimate hubness in datasets
"""


from __future__ import annotations

import logging
import warnings
from typing import Optional, Tuple, Union

import numpy as np
from scipy import stats
from tqdm.auto import tqdm

# Available hubness measures
VALID_HUBNESS_MEASURES = [
    "all",
    "all_but_gini",
    "k_skewness",
    "k_skewness_truncnorm",
    "atkinson",
    "gini",
    "robinhood",
    "antihubs",
    "antihub_occurrence",
    "hubs",
    "hub_occurrence",
    "groupie_ratio",
    "k_occurrence",
]


def _calc_skewness_truncnorm(k_occurrence: np.ndarray) -> float:
    """Hubness measure; corrected for non-negativity of k-occurrence.

    Hubness as skewness of truncated normal distribution
    estimated from k-occurrence histogram.

    Parameters
    ----------
    k_occurrence: ndarray
        Reverse nearest neighbor count for each object.
    Returns
    -------
    skew_truncnorm
    """
    clip_left = 0
    clip_right = np.iinfo(np.int64).max
    k_occurrence_mean = k_occurrence.mean()
    k_occurrence_std = k_occurrence.std(ddof=1)
    a = (clip_left - k_occurrence_mean) / k_occurrence_std
    b = (clip_right - k_occurrence_mean) / k_occurrence_std
    skew_truncnorm = stats.truncnorm(a, b).moment(3)
    return skew_truncnorm


def _calc_gini_index(
    k_occurrence: np.ndarray, limiting="memory", verbose: int = 0
) -> float:
    """Hubness measure; Gini index
    Parameters
    ----------
    k_occurrence: ndarray
        Reverse nearest neighbor count for each object.
    limiting: 'memory' or 'cpu'
        If 'cpu', use fast implementation with high memory usage,
        if 'memory', use slightly slower, but memory-efficient implementation,
        otherwise use naive implementation (slow, low memory usage)
    verbose: int
        control verbosity
    Returns
    -------
    gini_index
    """
    n = k_occurrence.size
    if limiting in ["memory", "space"]:
        numerator = 0
        for i in tqdm(range(n), disable=not verbose, desc="Gini"):
            numerator += np.sum(np.abs(k_occurrence[:] - k_occurrence[i]))
    elif limiting in ["time", "cpu"]:
        numerator = np.sum(
            np.abs(k_occurrence.reshape(1, -1) - k_occurrence.reshape(-1, 1))
        )
    else:  # slow naive implementation
        n = k_occurrence.size
        numerator = 0
        for i in range(n):
            for j in range(n):
                numerator += np.abs(k_occurrence[i] - k_occurrence[j])
    denominator = 2 * n * np.sum(k_occurrence)
    return numerator / denominator


def _calc_robinhood_index(k_occurrence: np.ndarray) -> float:
    """Hubness measure; Robin hood/Hoover/Schutz index.

    Parameters
    ----------
    k_occurrence: ndarray
        Reverse nearest neighbor count for each object.
    Returns
    -------
    robinhood_index

    Notes
    -----
    The Robin Hood index was proposed in [1]_ and is especially suited
    for hubness estimation in large data sets. Additionally, it offers
    straight-forward interpretability by answering the question:
    What share of k-occurrence must be redistributed, so that all objects
    are equally often nearest neighbors to others?

    References
    ----------
    .. [1] `Feldbauer, R.; Leodolter, M.; Plant, C. & Flexer, A.
            Fast approximate hubness reduction for large high-dimensional data.
            IEEE International Conference of Big Knowledge (2018).`
    """
    numerator = 0.5 * float(np.sum(np.abs(k_occurrence - k_occurrence.mean())))
    denominator = float(np.sum(k_occurrence))
    return numerator / denominator


def _calc_atkinson_index(k_occurrence: np.ndarray, eps: float = 0.5) -> float:
    """Hubness measure; Atkinson index.

    Parameters
    ----------
    k_occurrence: ndarray
        Reverse nearest neighbor count for each object.
    eps: float
        'Income' weight. Turns the index into a normative measure.
    Returns
    -------
    atkinson_index
    """
    if eps == 1:
        term = np.prod(k_occurrence) ** (1.0 / k_occurrence.size)
    else:
        term = np.mean(k_occurrence ** (1 - eps)) ** (1 / (1 - eps))
    return float(1.0 - 1.0 / k_occurrence.mean() * term)


def _calc_antihub_occurrence(k_occurrence: np.ndarray) -> Tuple[np.ndarray, float]:
    """Proportion of antihubs in data set.

    Antihubs are objects that are never among the nearest neighbors
    of other objects.

    Parameters
    ----------
    k_occurrence: ndarray
        Reverse nearest neighbor count for each object.
    Returns
    -------
    antihubs, antihub_occurrence
    """
    antihubs = np.argwhere(k_occurrence == 0).ravel()
    antihub_occurrence = antihubs.size / k_occurrence.size
    return antihubs, antihub_occurrence


def _calc_hub_occurrence(
    k: int, k_occurrence: np.ndarray, n_test: int, hub_size: float = 2
) -> Tuple[np.ndarray, float]:
    """Proportion of nearest neighbor slots occupied by hubs.

    Parameters
    ----------
    k: int
        Specifies the number of nearest neighbors
    k_occurrence: ndarray
        Reverse nearest neighbor count for each object.
    n_test: int
        Number of queries (or objects in a test set)
    hub_size: float
        Factor to determine hubs
    Returns
    -------
    hubs, hub_occurrence
    """
    hubs = np.argwhere(k_occurrence >= hub_size * k).ravel()
    hub_occurrence = k_occurrence[hubs].sum() / k / n_test
    return hubs, hub_occurrence


def hubness_score(
    nn_ind: np.ndarray,
    target_samples: int,
    *,
    k: Optional[int] = None,
    hub_size: float = 2.0,
    shuffle_equal: bool = True,
    random_state=None,
    verbose: int = 0,
    return_value: str = "all_but_gini",
    store_k_occurrence: bool = False,
) -> Union[float, dict]:
    """Calculates hubness scores from given neighbor indices

    Utilizes findings from [1]_ and [2]_.

    Parameters
    ----------
    nn_ind : np.ndarray
       Neighbor index matrix
    target_samples: int
        number of entities in the target space
    k : int
        number of k for k-nearest neighbor
    hub_size : float
        Hubs are defined as objects with k-occurrence > hub_size * k.
    shuffle_equal : bool
        If true shuffle neighbors with identical distances
        to avoid artifact hubness.
        NOTE: This is especially useful for secondary distance measures
        with a finite number of possible values
    random_state: int, RandomState instance or None, optional
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.
    verbose : int
        Level of output messages
    return_value : str
        Hubness measure to return
        By default, return all but gini,
        because gini is slow on large datasets
        Use "all" to return a dict of all available measures,
        or check `kiez.analysis.VALID_HUBNESS_MEASURE`
        for available measures.
    store_k_occurrence: bool
        Whether to save the k-occurrence. Requires O(n_test) memory.

    Returns
    -------
    hubness_measure: float or dict
            Return the hubness measure as indicated by `return_value`.
            if return_value is 'all',
            a dict of all hubness measures is returned.

    Raises
    ------
    ValueError
        If nn_ind has wrong type

    References
    ----------
    .. [1] `Radovanović, M.; Nanopoulos, A. & Ivanovic, M.
            Hubs in space: Popular nearest neighbors in high-dimensional data.
            Journal of Machine Learning Research, 2010, 11, 2487-2531`
    .. [2] `Feldbauer, R.; Leodolter, M.; Plant, C. & Flexer, A.
            Fast approximate hubness reduction for large high-dimensional data.
            IEEE International Conference of Big Knowledge (2018).`

    Examples
    --------
    >>> from kiez import Kiez
    >>> from kiez.analysis import hubness_score
    >>> import numpy as np
    >>> # create example data
    >>> rng = np.random.RandomState(0)
    >>> source = rng.rand(100,50)
    >>> target = rng.rand(100,50)
    >>> # fit and get neighbors
    >>> k_inst = Kiez()
    >>> k_inst.fit(source, target)
    >>> nn_ind = k_inst.kneighbors(return_distance=False)
    >>> # get hubness
    >>> hub_score = hubness_score(nn_ind, target.shape[1])
    >>> hub_score["robinhood"]
        0.31
    """
    n_train = nn_ind.shape[0]
    n_test = target_samples
    k_neighbors = nn_ind.copy()
    if k is None:
        k = nn_ind.shape[1]
    else:
        if k < k_neighbors.shape[1]:
            k_neighbors = k_neighbors[:, :k]
        elif k > k_neighbors.shape[1]:
            k = nn_ind.shape[1]
            warnings.warn(f"k > nn_ind.shape[1], k will be set to {k}")
    assert k is not None

    # Negative indices can occur, when ANN does not find enough neighbors,
    # and must be removed
    mask = k_neighbors < 0
    if np.any(mask):
        k_neighbors = k_neighbors[~mask]
        del mask

    try:
        k_occurrence = np.bincount(k_neighbors.astype(int).ravel(), minlength=n_train)
    except ValueError as e:
        logging.info(f"k_occurence failed with the following neighbors: {k_neighbors}")
        raise e

    # traditional skewness measure
    k_skewness = stats.skew(k_occurrence)

    # new skewness measure (truncated normal distribution)
    k_skewness_truncnorm = _calc_skewness_truncnorm(k_occurrence)

    # Gini index
    if return_value in ["gini", "all"]:
        limiting = "space" if k_occurrence.shape[0] > 10_000 else "time"
        gini_index = _calc_gini_index(k_occurrence, limiting, verbose=verbose)
    else:
        gini_index = np.nan

    # Robin Hood index
    robinhood_index = _calc_robinhood_index(k_occurrence)

    # Atkinson index
    atkinson_index = _calc_atkinson_index(k_occurrence)

    # anti-hub occurrence
    antihubs, antihub_occurrence = _calc_antihub_occurrence(k_occurrence)

    # hub occurrence
    hubs, hub_occurrence = _calc_hub_occurrence(
        k=k,
        k_occurrence=k_occurrence,
        n_test=n_test,
        hub_size=hub_size,
    )

    # Largest hub
    groupie_ratio = k_occurrence.max() / n_test / k

    # Dictionary of all hubness measures
    hubness_measures = {
        "k_skewness": k_skewness,
        "k_skewness_truncnorm": k_skewness_truncnorm,
        "atkinson": atkinson_index,
        "gini": gini_index,
        "robinhood": robinhood_index,
        "antihubs": antihubs,
        "antihub_occurrence": antihub_occurrence,
        "hubs": hubs,
        "hub_occurrence": hub_occurrence,
        "groupie_ratio": groupie_ratio,
    }

    if store_k_occurrence:
        hubness_measures["k_occurrence"] = k_occurrence
    if return_value == "all":
        return hubness_measures
    elif return_value == "all_but_gini":
        del hubness_measures["gini"]
        return hubness_measures
    else:
        return hubness_measures[return_value]
